/*
 * Copyright (c) 2002 Akihiko GOTANDA.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#if !defined(USE_SYSTEM)
#include <sys/wait.h>
#endif
#include <ctype.h>
#include <err.h>
#include <ifaddrs.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#define SYSLOG_NAMES
#include <syslog.h>
#undef SYSLOG_NAMES
#include <net/if.h>
#include <net/if_media.h>
#if !defined(WITHOUT_ADDR)
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#ifndef lint
static const char rcs_id[] = "$Id: nicmond.c,v 1.11 2002/06/09 02:05:00 a-gota Exp $";
#endif

#define MAXSTRLEN 1024

#if !defined(WITHOUT_ADDR)
char addr_buf[MAXHOSTNAMELEN + 1];
#endif
char str_tmp[MAXSTRLEN];

char *conffile;

#define DEFAULT_CONF "nicmond.conf"
u_int interval = 5;
u_int resense = 2;
int facility = LOG_DAEMON;
int log_fatal = LOG_EMERG;
int log_critical = LOG_CRIT;
int log_notice = LOG_NOTICE;
int log_info = LOG_INFO;

int debug_f = 0;
int verbose_f = 0;
int reload_f = 0;
int forcedown_f[2] = { 0, 0 };

typedef struct {
    const char *name;
    void *value_p;
    int type;
    int (*func)(char *str);
} param_t;

#define PARAM_T_NONE	0
#define PARAM_T_UINT	1
#define PARAM_T_UINTZ	2
#define PARAM_T_INT	3
#define PARAM_T_STR	4
#define PARAM_T_FUNC	5

static int conf_get_facility(char *str);
static int conf_get_priority(char *str);

param_t paramlist[] = {
    { "interval",	&interval,	PARAM_T_UINT,	NULL },
    { "resense",	&resense,	PARAM_T_UINTZ,	NULL },
    { "facility",	&facility,	PARAM_T_FUNC,	conf_get_facility },
    { "fatal",		&log_fatal,	PARAM_T_FUNC,	conf_get_priority },
    { "critical",	&log_critical,	PARAM_T_FUNC,	conf_get_priority },
    { "notice",		&log_notice,	PARAM_T_FUNC,	conf_get_priority },
    { "info",		&log_info,	PARAM_T_FUNC,	conf_get_priority },
    { NULL,		NULL,		PARAM_T_NONE,	NULL }
};


static void reloadreq(int unused)
{
    reload_f = 1;
    if (debug_f) {
	warnx("SIGHUP received.");
    }
}


static void termreq(int unused)
{
    if (debug_f) {
	warnx("SIGTERM received.");
    } else if (log_notice != INTERNAL_NOPRI) {
	syslog(log_notice, "terminated normally");
    }
    closelog();

    exit (0);
}


static void downreq(int signal)
{
    if (signal == SIGUSR1) {
	if (debug_f) {
	    warnx("SIGUSR1 received.");
	} else if (log_notice != INTERNAL_NOPRI) {
	    syslog(log_notice, "SIGUSR1 received");
	}
	forcedown_f[0] = 1;
    } else {
	if (debug_f) {
	    warnx("SIGUSR2 received.");
	} else if (log_notice != INTERNAL_NOPRI) {
	    syslog(log_notice, "SIGUSR2 received");
	}
	forcedown_f[1] = 1;
    }
}


static void usage(char *argv0)
{
    fprintf(stderr, "Usage %s [options] <if> [<if> ...]\n", argv0);
    fprintf(stderr, "\toptions: -a        don't remove no exist interfaces from watch list\n");
    fprintf(stderr, "\t         -d        debug mode\n");
    fprintf(stderr, "\t         -d -d     dummy mode\n");
    fprintf(stderr, "\t         -f <conf file>\n");
    fprintf(stderr, "\t                   specifies configration file\n");
    fprintf(stderr, "\t                   default: \"%s\"\n", DEFAULT_CONF);
    fprintf(stderr, "\t         -i <interval>\n");
    fprintf(stderr, "\t                   interval time for polling I/F (unit: sec.)\n");
    fprintf(stderr, "\t         -p <pid file>\n");
    fprintf(stderr, "\t                   specifies filename to write PID\n");
    fprintf(stderr, "\t                   default: none\n");
    fprintf(stderr, "\t         -s {0|1}  internal initial state for all interface(s)\n");
    fprintf(stderr, "\t         -v:       verbose mode\n");
    fprintf(stderr, "\tif:      <ifname>[:<initial state>]\n");
    fprintf(stderr, "\t                   internal initial state (0: down, 1: up)\n");
    fprintf(stderr, "\texample: %s -f /usr/local/etc/%s dc0 fxp0:1\n",
	    argv0, DEFAULT_CONF);
}


int main(int argc, char *argv[])
{
    int cnt, addrcount, prevstat, num_valid;
    FILE *conffp = NULL, *pidfp = NULL;
    int initstat = 0;
    int *stat, *bad_f, allow_noexist_f = 0;
    struct sockaddr_storage *sa_p;
    int ch, tmp, errors;
    char *ptr, *argv0;

    static void conf_get_opt(FILE *fp);
    static int check_status(char *if_name, int bad_f);
#if !defined(WITHOUT_ADDR)
    static int get_assigned_addr(const char *if_name, struct sockaddr_storage **sa);
#endif
    static int conf_find_interface(FILE *conffp, char *if_name);
    static int process_interface(FILE *conffp,
				 char *if_name, int stat, int addrcount,
				 struct sockaddr_storage *sa);

    argv0 = argv[0];

    errors = 0;
    while ((ch = getopt(argc, argv, "adf:i:p:s:v")) != EOF) {
	switch (ch) {
	  case 'a':
	    ++allow_noexist_f;
	    break;
	  case 'd':
	    ++debug_f;
	    break;
	  case 'f':
	    if ((conffp = fopen(optarg, "r")) == NULL) {
		warn("%s", optarg);
	    } else {
		conffile = strdup(optarg);
	    }
	    break;
	  case 'i':
	    if ((tmp = atoi(optarg)) <= 0) {
		warnx("invalid interval time (%d)", tmp);
	    }
	    break;
	  case 'p':
	    if ((pidfp = fopen(optarg, "w")) == NULL) {
		warn("%s", optarg);
	    }
	    break;
	  case 's':
	    tmp = atoi(optarg);
	    if (tmp < 0 || 1 < tmp) {
		warnx("invalid initial state (%d)", tmp);
	    } else {
		initstat = tmp;
	    }
	    break;
	  case 'v':
	    ++verbose_f;
	    break;
	  default:
	    warnx("'%c': unknown option", ch);
	    ++errors;
	    break;
	}
    }
    argc -= optind;
    argv += optind;

    if (argc < 1 || errors) {
	usage(argv0);
	return 1;
    }

    if (debug_f == 1) {
	warnx("running under debug mode.");
    } else if (1 < debug_f) {
	warnx("running under dummy mode.");
    }

    if (conffp == NULL) {
	if ((conffp = fopen(DEFAULT_CONF, "r")) == NULL) {
	    err(1, "%s", DEFAULT_CONF);
	} else {
	    conffile = strdup(DEFAULT_CONF);
	}
    }

    conf_get_opt(conffp);

    num_valid = argc;
    if ((stat = calloc(num_valid, sizeof(int))) == NULL) {
	errx(1, NULL);
    }
    if ((bad_f = calloc(num_valid, sizeof(int))) == NULL) {
	errx(1, NULL);
    }

    for (cnt = 0; cnt < argc; cnt++) {
	if ((ptr = strchr(argv[cnt], ':')) != NULL) {
	    *ptr++ = '\0';
	    if ((tmp = atoi(ptr)) < 0 || 1 < tmp) {
		warnx("invalid initial state (%d)", tmp);
	    } else {	
		stat[cnt] = tmp;
	    }
	} else {
	    stat[cnt] = initstat;
	}
	bad_f[cnt] = 0;
    }

    if (!debug_f) {
	if (daemon(0, 1)) {
	    warn(NULL);
	    errx(1, "fork failed");
	}
    }

    if (pidfp != NULL) {
	fprintf(pidfp, "%d\n", getpid());
	fclose(pidfp);
	pidfp = NULL;
    }

    signal(SIGHUP, reloadreq);
    signal(SIGTERM, termreq);
    signal(SIGUSR1, downreq);
    signal(SIGUSR2, downreq);

    if (!debug_f) {
	if ((ptr = strrchr(argv0, '/')) != NULL) {
	    argv0 = ptr + 1;
	}
	if (facility != INTERNAL_MARK) {
	    openlog(argv0, LOG_CONS | LOG_PID, facility);
	}
	if (log_notice != INTERNAL_NOPRI) {
	    syslog(log_notice, "started");
	}
    }

    errors = 0;
    do {
	for (cnt = 0; cnt < argc; cnt++) {
	    prevstat = stat[cnt];
	    if (argv[cnt][0] == '\0') {
		continue;
	    }
	    if (cnt < 2 && 0 < forcedown_f[cnt]) {
		if (debug_f) {
		    warnx("force down %s", argv[cnt]);
		} else if (log_notice != INTERNAL_NOPRI) {
		    syslog(log_notice, "force down %s", argv[cnt]);
		}
		stat[cnt] = 0;
		forcedown_f[cnt] = -1;
		goto skip;
	    }

	    if ((tmp = check_status(argv[cnt], bad_f[cnt])) < 0) {
		if (tmp < -10) {
		    if (!bad_f[cnt]) {
			if (!debug_f && log_critical != INTERNAL_NOPRI) {
			    syslog(log_critical,
				   "can't watch for %s; ignored", argv[cnt]);
			}
		    }
		    if (!allow_noexist_f) {
			argv[cnt][0] = '\0';
			--num_valid;
		    }
		}
		bad_f[cnt] = 1;
		continue;
	    } else if (bad_f[cnt]) {
		if (!debug_f && log_critical != INTERNAL_NOPRI) {
		    syslog(log_critical, "%s is alive", argv[cnt]);
		}
		bad_f[cnt] = 0;
	    }

	    if (resense != 0 && tmp != prevstat) {
		int tmp2;

		sleep(resense);
		if ((tmp2 = check_status(argv[cnt], 0)) != tmp) {
		    sleep(resense);
		    if (check_status(argv[cnt], 0) != tmp2) {
			if (verbose_f) {
			    warnx("%s: link status is unstable", argv[cnt]);
			}
			if (!debug_f && log_fatal != INTERNAL_NOPRI) {
			    syslog(log_fatal,
				   "%s: link status is unstable", argv[cnt]);
			}
			continue;
		    } else {
			tmp = tmp2;
		    }
		}
	    }

	    if (cnt < 2 && forcedown_f[cnt] < 0) {
		if (tmp == 0) {
		    if (debug_f) {
			warnx("%s: down delected, resume to normal mode",
			      argv[cnt]);
		    } else if (log_notice != INTERNAL_NOPRI) {
			syslog(log_notice,
			       "%s: down delected, resume to normal mode",
			       argv[cnt]);
		    }
		    forcedown_f[cnt] = 0;
		}
		continue;
	    }

	    stat[cnt] = tmp;
	skip:
	    if (stat[cnt] != prevstat) {
		if (verbose_f) {
		    printf("%s: status changed\n", argv[cnt]);
		}
		if (!debug_f && log_notice != INTERNAL_NOPRI) {
		    syslog(log_notice,
			   "%s: link status changed (%d->%d)",
			   argv[cnt], prevstat, stat[cnt]);
		}
#if defined(WITHOUT_ADDR)
		addrcount = 0;
		sa_p = NULL;
#else
		addrcount = get_assigned_addr(argv[cnt], &sa_p);
		if (verbose_f) {
		    printf("%s has %d address(es)\n", argv[cnt], addrcount);
		}
#endif
		if (conf_find_interface(conffp, argv[cnt])) {
		    process_interface(conffp, argv[cnt], stat[cnt], addrcount, sa_p);
		    if (!debug_f && log_notice != INTERNAL_NOPRI) {
			syslog(log_notice,
			       "%s: all action(s) done", argv[cnt]);
		    }
		}
#if !defined(WITHOUT_ADDR)
		if (sa_p != NULL) {
		    free(sa_p);
		}
#endif
	    }
	}
	if (num_valid < 1) {
	    ++errors;
	    if (!debug_f && log_fatal != INTERNAL_NOPRI) {
		syslog(log_fatal, "no valid interface");
	    }
	} else {
	    sleep(interval);

	    if (reload_f) {
		fclose(conffp);
		if ((conffp = fopen(conffile, "r")) == NULL) {
		    warn("%s", conffile);
		    if (!strcmp(conffile, DEFAULT_CONF)) {
			free(conffile);
			return 1;
		    }
		    free(conffile);
		    if ((conffp = fopen(DEFAULT_CONF, "r")) == NULL) {
			err(1, "%s", DEFAULT_CONF);
		    } else {
			conffile = strdup(DEFAULT_CONF);
		    }
		}

		if (debug_f) {
		    warnx("Reloading configuration from \"%s\"", conffile);
		} else {
		    if (log_notice != INTERNAL_NOPRI) {
			syslog(log_notice,
			   "Reloading configuration from \"%s\"", conffile);
		    }
		}
		tmp = facility;
		conf_get_opt(conffp);
		reload_f = 0;

		if (!debug_f && facility != tmp) {
		    if (log_notice != INTERNAL_NOPRI) {
			syslog(log_notice,
			   "changed facility from %d to %d", tmp, facility);
		    }
		    closelog();
		    if (facility != INTERNAL_MARK) {
			openlog(argv0, LOG_CONS | LOG_PID, facility);
		    }
		}
	    }
	}
    } while (!errors);

    if (!debug_f && log_notice != INTERNAL_NOPRI) {
	syslog(log_notice, "terminated abnormally");
	closelog();
    }

    return 1;
}


static int check_status(char *if_name, int bad_f)
{
    int s, ret;
    struct ifmediareq ifmr;

    if (if_nametoindex(if_name) == 0) {
	if (!bad_f) {
	    if (debug_f) {
		warnx("%s: no such a interface", if_name);
	    } else {
		syslog(log_critical, "%s: no such a interface", if_name);
	    }
	}
	return -11;
    }

    if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	if (!bad_f) {
	    if (debug_f) {
		warnx("%s: cannot open socket", if_name);
	    } else {
		syslog(log_critical, "%s: cannot open socket", if_name);
	    }
	}
	return -1;
    }

    memset(&ifmr, 0, sizeof(ifmr));
    strncpy(ifmr.ifm_name, if_name, sizeof(ifmr.ifm_name));
    ret = ioctl(s, SIOCGIFMEDIA, (caddr_t)&ifmr);
    close(s);
    if (ret < 0) {
	if (!bad_f) {
	    if (debug_f) {
		warnx("%s: interface doesn't support SIOC{G,S}IFMEDIA",
		      if_name);
	    } else {
		syslog(log_critical,
		       "%s: interface doesn't support SIOC{G,S}IFMEDIA", if_name);
	    }
	}
	return -12;
    }

    if (ifmr.ifm_status & IFM_AVALID) {
	if (ifmr.ifm_status & IFM_ACTIVE) {
	    if (debug_f || verbose_f) {
		printf("%s: active\n", if_name);
	    }
	    return 1;
	} else {
	    if (debug_f || verbose_f) {
		printf("%s: inactive\n", if_name);
	    }
	    return 0;
	}
    } else {
	if (!bad_f) {
	    if (debug_f) {
		warnx("%s: invalid", if_name);
	    } else {
		syslog(log_critical, "%s: invalid", if_name);
	    }
	}
	return -2;
    }
}


#if !defined(WITHOUT_ADDR)
static int get_assigned_addr(const char *if_name, struct sockaddr_storage **addr)
{
    int addrcount;
    struct ifaddrs *ifa, *ifap;
    struct sockaddr_storage *ptr;
    struct sockaddr_in6 *sin6;

    if (getifaddrs(&ifa) != 0) {
	warn("%s: %s: ", if_name, "getifaddrs");
	*addr = NULL;
	return 0;
    }
    addrcount = 0;
    for (ifap = ifa; ifap != NULL; ifap = ifap->ifa_next) {
	if (strcmp(if_name, ifap->ifa_name)) {
	    continue;
	}
	if (ifap->ifa_addr->sa_family == AF_INET6) {
	    sin6 = (struct sockaddr_in6 *)ifap->ifa_addr;
	    if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr)
		&& *(u_short *)&sin6->sin6_addr.s6_addr[2] != 0) {
		/* skip link-local address */
		continue;
	    }
	} else if (ifap->ifa_addr->sa_family != AF_INET) {
	    continue;
	}
	++addrcount;
    }

    if ((*addr = calloc(addrcount, sizeof(struct sockaddr_storage))) == NULL) {
	warn("%s: get_assigned_addr: ", if_name);
	return 0;
    }

    ptr = *addr;
    for (ifap = ifa; ifap != NULL; ifap = ifap->ifa_next) {
	if (strcmp(if_name, ifap->ifa_name)) {
	    continue;
	}
	if (ifap->ifa_addr->sa_family == AF_INET6) {
	    sin6 = (struct sockaddr_in6 *)ifap->ifa_addr;
	    if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr)
		&& *(u_short *)&sin6->sin6_addr.s6_addr[2] != 0) {
		/* skip link-local address */
		continue;
	    }
	} else if (ifap->ifa_addr->sa_family != AF_INET) {
	    continue;
	}
	memcpy(ptr++, ifap->ifa_addr, ifap->ifa_addr->sa_len);
    }

    freeifaddrs(ifap);

    return addrcount;
}
#endif


static int process_interface(FILE *conffp,
			     char *if_name, int stat,
			     int addrcount, struct sockaddr_storage *ss)
{
#if !defined(WITHOUT_ADDR)
    struct sockaddr_storage *ssp;
    struct sockaddr_in *sin;
    struct sockaddr_in6 *sin6;
    const char *addr;
    char *ptr;
    int cnt;
#endif
    int eof_f;
    char *action;
    char cmdbuf[2][MAXSTRLEN], *cmd = NULL;

    static char *conf_find_action(FILE *conffp, int stat, int *eof_f);
    static char *replace_string(char *dst, char *src, char *str, const char *rep);
    static int exec_action(char *action, FILE *conffp);

    eof_f = 0;
    while (!eof_f && (action = conf_find_action(conffp, stat, &eof_f)) != NULL) {
	strcpy(cmdbuf[0], action);
	replace_string(cmdbuf[1], cmdbuf[0], "$device", if_name);
	cmd = cmdbuf[1];

#if !defined(WITHOUT_ADDR)
	if (((ptr = strstr(cmdbuf[1], "$myaddr4")) == NULL || ptr[-1] == '\\')
	    && ((ptr = strstr(cmdbuf[1], "$myaddr6")) == NULL || ptr[-1] == '\\')) {
#endif
	    /* not contain address part */
	    exec_action(cmdbuf[1], conffp);
	    continue;
#if !defined(WITHOUT_ADDR)
	}

	ssp = ss;
	for (cnt = 0; cnt < addrcount; cnt++) {
	    addr = NULL;
	    switch (ssp->ss_family) {
	      case AF_INET:
		sin = (struct sockaddr_in *)ssp;
		addr = (char *)inet_ntoa(sin->sin_addr);
		if (verbose_f) {
		    printf("IPv4: %s\n", addr);
		}
		break;
	      case AF_INET6:
		sin6 = (struct sockaddr_in6 *)ssp;
		addr = inet_ntop(AF_INET6, &sin6->sin6_addr,
				 addr_buf, sizeof(addr_buf));
		if (verbose_f) {
		    printf("IPv6: %s\n", addr);
		}
		break;
	      default:
		break;
	    }

	    if (addr != NULL) {
		if (strstr(cmdbuf[1], "$myaddr6") != NULL
			&& ssp->ss_family == AF_INET6) {
		    exec_action(replace_string(cmdbuf[0], cmdbuf[1],
					       "$myaddr6", addr),
				conffp);
		} else if (strstr(cmdbuf[1], "$myaddr4") != NULL
			   && ssp->ss_family == AF_INET) {
		    exec_action(replace_string(cmdbuf[0], cmdbuf[1],
					       "$myaddr4", addr),
				conffp);
		}
	    }

	    ++ssp;
	}
#endif
    }

    return 0;
}


char *skip_ws(char *ptr)
{
    while (isspace(*ptr)) {
	++ptr;
    }
    return ptr;
}


static char *conf_get_line(FILE *conffp, int *eof_f)
{
    int too_long;
    static char buf[MAXSTRLEN];
    char *ptr, *ret = NULL;

    *eof_f = 0;
    if (!*eof_f && (ret = fgets(buf, sizeof(buf), conffp)) != NULL) {
	too_long = 0;
	if ((ptr = strchr(buf, '\n')) == NULL) {
	    too_long = 1;	/* line is too long */
	} else {
	    *ptr = '\0';
	}

	/* remove comment */
	ptr = buf;
	while ((ptr = strchr(ptr, '#')) != NULL) {
	    char *src, *dst;

	    if (ptr == buf || ptr[-1] != '\\') {
		*ptr = '\0';
		break;
	    } else {
		dst = (src = ptr++) -1;
		while ((*dst++ = *src++)) {
		    ;
		}
	    }
	}

	/* remove trailing white spaces */
	ptr = &buf[strlen(buf) - 1];
	while (buf <= ptr && isspace(*ptr)) {
	    *ptr-- = '\0';
	}

	if (verbose_f) {
	    warnx("read conf: %s", buf);
	}

	if (too_long) {
	    /* skip trailing garbage */
	    while ((ptr = fgets(str_tmp, sizeof(str_tmp), conffp)) != NULL) {
		if (strchr(str_tmp, '\n') != NULL) {
		    break;
		}
	    }
	    if (ptr == NULL) {
		/* EOF */
		*eof_f = 1;
	    }
	}
    }

    if (ret == NULL) {
	/* EOF */
	*eof_f = 1;
    }

    if (*eof_f) {
	return NULL;
    }
    return buf;
}


static char *conf_wordmatch(char *buf, const char *str)
{
    int len;
    char *res = NULL;

    if (!strncmp(buf, str, (len = strlen(str)))) {
	res = buf + len;
	if (*res != '\0') {
	    if (!isspace(*res)) {
		res = NULL;
	    } else {
		res = skip_ws(res);
	    }
	}
    }

    return res;
}


static void conf_get_opt(FILE *fp)
{
    int eof_f;
    int tmp;
    char *line, *ptr;
    param_t *param_p;

    rewind(fp);
    while ((line = conf_get_line(fp, &eof_f)) != NULL) {
	for (param_p = paramlist; param_p->name != NULL; param_p++) {
	    if ((ptr = conf_wordmatch(line, param_p->name)) != NULL) {
		switch (param_p->type) {
		  case PARAM_T_UINT:
		    if ((tmp = atoi(ptr)) <= 0) {
			warnx("invalid %s value (%d)", param_p->name, tmp);
			break;
		    }
		    *(int *)param_p->value_p = tmp;
		    break;
		  case PARAM_T_UINTZ:
		    if ((tmp = atoi(ptr)) < 0) {
			warnx("invalid %s value (%d)", param_p->name, tmp);
			break;
		    }
		    *(int *)param_p->value_p = tmp;
		    break;
		  case PARAM_T_INT:
		    *(int *)param_p->value_p = atoi(ptr);
		    break;
		  case PARAM_T_STR:
		    tmp = (int)param_p->func;	/* get max length */
		    strncpy(param_p->value_p, ptr, tmp);
		    ((char *)param_p->value_p)[tmp - 1] = '\0';
		    break;
		  case PARAM_T_FUNC:
		    if (0 <= (tmp = param_p->func(ptr))) {
			*(int *)param_p->value_p = tmp;
		    }
		    break;
		  default:
		    warnx("BUG! invalid parameter type assigned (%d)",
			  param_p->type);
		    break;
		}
		break;
	    }
	}

	if (eof_f) {
	    /* EOF */
	    break;
	}
    }

    rewind(fp);
}


static int conf_get_facility(char *str)
{
    CODE *code_p;

    if (!strcmp(str, "none")) {
	return INTERNAL_MARK;
    }
    for (code_p = facilitynames; code_p->c_name != NULL; code_p++){
	if (!strcmp(code_p->c_name, str)) {
	    return  code_p->c_val;
	}
    }
    warnx("invalid facirity name (%s)", str);

    return -1;
}


static int conf_get_priority(char *str)
{
    CODE *code_p;

    for (code_p = prioritynames; code_p->c_name != NULL; code_p++){
	if (!strcmp(code_p->c_name, str)) {
	    return  code_p->c_val;
	}
    }
    warnx("invalid priority name (%s)", str);

    return -1;
}


static char *conf_find_entry(FILE *conffp, char *str, int *eof_f)
{
    char *ptr, *res = NULL;

    while (res == NULL && (ptr = conf_get_line(conffp, eof_f)) != NULL) {
	res = conf_wordmatch(ptr, str);

	if (*eof_f) {
	    /* EOF */
	    break;
	}
    }

    return res;
}


static int conf_find_interface(FILE *conffp, char *if_name)
{
    int eof_f;
    char *ptr = NULL;

    rewind(conffp);
    eof_f = 0;
    while (!eof_f && (ptr = conf_find_entry(conffp, "interface", &eof_f)) != NULL) {
	if (conf_wordmatch(ptr, if_name) != NULL && !eof_f) {
	    if (verbose_f) {
		warnx("find_interface: %s found", if_name);
	    }
	    return 1;	/* found */
	}
    }

    return 0;	/* EOF -- not found */
}


static char *conf_find_action(FILE *conffp, int stat, int *eof_f)
{
    char *ptr = NULL;

    *eof_f = 0;
    while (!*eof_f && (ptr = conf_get_line(conffp, eof_f)) != NULL) {
	if (*ptr == '\0') {
	    /* ignore empty line */
	    continue;
	}
	if (*ptr != '\t') {
	    /* end of section */
	    break;
	}
	ptr = skip_ws(ptr);
	if (stat == 1 && (ptr = conf_wordmatch(ptr, "up")) != NULL) {
	    if (verbose_f) {
		warnx("find_action: UP found: %s", ptr);
	    }
	    return ptr;
	}
	if (stat == 0 && (ptr = conf_wordmatch(ptr, "down")) != NULL) {
	    if (verbose_f) {
		warnx("find_action: DOWN found: %s", ptr);
	    }
	    return ptr;
	}
    }

    if (ptr == NULL) {
	*eof_f = 1;
    }

    return NULL;
}


static char *replace_string(char *dst, char *src, char *str, const char *rep)
{
    char *srcp, *dstp, *ptr;
    const char *rep_p;
    int len, matched;

    len = strlen(str);
    srcp = src;
    dstp = dst;
    matched = 0;
    while ((ptr = strstr(srcp, str)) != NULL) {
	if (ptr != src && ptr[-1] == '\\') {
	    int cnt;

	    ptr[-1]  = '\0';
	    while ((*dstp++ = *srcp++)) {
		;
	    }
	    --dstp;
	    for (cnt = 0; cnt < len; cnt++) {
		*dstp++ = *srcp++;
	    }
	    continue;
	}

	++matched;
	if (rep == NULL) {
	    return NULL;
	}
	*ptr = '\0';
	while ((*dstp++ = *srcp++)) {
	    ;
	}
	srcp = ptr + len;
	--dstp;
	rep_p = rep;
	while ((*dstp++ = *rep_p++)) {
	    ;
	}
	--dstp;
    }
    while ((*dstp++ = *srcp++)) {
	;
    }

    return dst;
}


static int exec_action(char *cmd, FILE *conffp)
{
#if !defined(USE_SYSTEM)
    char **argv, *ptr;
    int argc, cnt;
    pid_t pid;
    int status;
#endif

    if (debug_f) {
	warnx("exec: %s", cmd);
	if (1 < debug_f) {
	    /* dummy */
	    return 0;
	}
    } else if (log_info != INTERNAL_NOPRI) {
	syslog(log_info, "exec: %s", cmd);
    }

#if defined(USE_SYSTEM)
    system(cmd);
#else
    ptr = cmd;
    for (cnt = 1; (ptr = strchr(ptr, ' ')) != NULL; cnt++) {
	++ptr;
    }
    if (debug_f) {
	warnx("%d argument(s)", cnt);
    }
    argv = (char **)calloc(sizeof(char *), cnt);

    argv[0] = ptr = skip_ws(cmd);
    argc = 1;
    while ((ptr = strchr(ptr, ' ')) != NULL) {
	*ptr++ = '\0';
	ptr = skip_ws(ptr);
	argv[argc++] = ptr;
    }
    argv[argc] = NULL;

    if (debug_f) {
	fputs("exec:", stderr);
	for (cnt = 0; cnt < argc; cnt++) {
	    fprintf(stderr, " \"%s\"", argv[cnt]);
	}
	fputc('\n', stderr);
    }

    if ((pid = fork()) < 0) {
	/* fork failed */
	warn(NULL);
	free(argv);
	return -1;
    } else if (pid == 0) {
	/* child process */
	fclose(conffp);
	if (execvp(argv[0], argv)) {
	    err(-1, NULL);
	}
    } else {
	/* parent process */
	if (waitpid(pid, &status, 0) < 0) {
	    warn(NULL);
	    free(argv);
	    return -1;
	}
    }

    free(argv);
    return WEXITSTATUS(status);
#endif

    return 0;
}
